from os import environ
from pathlib import Path

from dotenv import load_dotenv

PROJECT_ROOT = Path("__file__").resolve().parent
ENV_PATH = PROJECT_ROOT / ".env"

load_dotenv(ENV_PATH)

ckip_dir = environ.get("CKIP_DIR")
ckip_path = environ.get("CKIP_PATH")
nlp_model = environ.get("NLP_MODEL")

CKIP_DIR = PROJECT_ROOT / ".ckip" if ckip_dir is None else ckip_dir
CKIP_PATH = CKIP_DIR / "ckip_drivers.pickle" if ckip_path is None else ckip_path
NLP_MODEL = "bert-base" if nlp_model is None else nlp_model



def check_en_or_zh_chars(sentence: str) -> bool:
    """The check_en_or_zh_chars method checks whether the sentece is a English or Mandarin sentence.

    Args:
        sentence (str)
    Returns:
        True if the sentence is English or Mandarin, False otherwise.
    """
    return all(
        map(
            lambda value: not (19968 <= ord(value) <= 40869)
            and not (97 <= ord(value) <= 122)
            and not (65 <= ord(value) <= 90),
            sentence,
        )
    )


def validate_sentence(sentence: str) -> str | None:
    """The validate_sentence method validates whether the sentence is English or Mandarin.

    Args:
        sentence (str)
    Returns:
        a str if the sentence is English or Mandarin, None otherwise.
    """
    stripped_sentence = sentence.strip()

    if not check_en_or_zh_chars(stripped_sentence) and not stripped_sentence.startswith(
        "◆ From:"
    ):
        return stripped_sentence

    return None


def filter_body(sentence: str) -> list[str]:
    """The filter_body function filters the post body that are not English or Mandarin sentences.

    Args:
        sentence (str)
    Retunrs:
        a list of strings
    """
    output = map(validate_sentence, sentence.split("\n"))
    return list(filter(lambda value: value is not None, output))



import asyncio
import xml.etree.ElementTree as ET


async def create_metadata_tag(name: str, value: str, tei_header: ET.Element) -> None:
    """The create_metadata_tag function adds a new metadata tag to the given `tei_header`.

    Args:
        name (str): the name of the metadata tag to add
        value (str): the value of the metadata tag to add
        tei_header (xml.etree.ElementTree.Element): a tei header
    """
    ET.SubElement(tei_header, "metadata", name=name).text = value


async def create_header_tag(root: ET.Element, meta_data: dict[str, str]) -> None:
    """The create_header_tag function creates a set of metadata tags and adds
    them to the `tei_header`.

    Args:
        root (xml.etree.ElementTree.Element): the root element
        meta_data (dict): the metadata of a post
    """
    tei_header = ET.SubElement(root, "teiHeader")
    tasks = []

    for name, value in meta_data.items():
        task = asyncio.create_task(create_metadata_tag(name, value, tei_header))
        tasks.append(task)

    await asyncio.gather(*tasks)



from abc import (
    ABC,
    abstractmethod,
)
import asyncio
from dataclasses import dataclass
import xml.etree.ElementTree as ET

from .cleaners import Cleaner

SegmentedSentences = list[list[tuple[str, str]]]
TagData = tuple[str, str] | list[tuple[str, str]] | dict[str, list[tuple[str, str]]]


@dataclass
class TeiTagger(ABC):
    """
    The TeiTagger object creats the tei tags.
    """

    tag: ET.Element
    data_cleaner: Cleaner

    @abstractmethod
    async def build_tag(self, tag_data: TagData):  # noqa
        pass

    async def build_tags(self, segmented_sentences: SegmentedSentences):
        tasks = []
        for segmented_sentence in segmented_sentences:
            task = asyncio.create_task(self.build_tag(segmented_sentence))
            tasks.append(task)

        return await asyncio.gather(*tasks)

    async def create(self):
        """The create method build TEI tags for each segmented sentence."""
        segmented_sentences = self.data_cleaner.clean()

        if not segmented_sentences:
            return None

        return await self.build_tags(segmented_sentences)



import asyncio
from typing import Literal
from xml.dom import minidom
import xml.etree.ElementTree as ET

from .tags.body import create_body_tags
from .tags.head import create_header_tag
from .utils import get_year

PostData = dict[str, str | dict[str, int] | list[dict[str, str]]]
Media = Literal["ptt", "dcard"]


async def create_tags(root: ET.Element, post_data: PostData, media: Media) -> None:
    """The create_tags function creates TEI XML tags and add them to the root.

    Args:
        root (xml.etree.ElementTree.Element): the root element of the TEI XML tree
        post_data (PostData): the post data
        media (Media): a Taiwan social media name
    """
    meta_data = {
        "media": media,
        "author": post_data.get("author"),
        "post_id": post_data.get("post_id"),
        "year": get_year(post_data.get("date")),
        "board": post_data.get("board"),
        "title": post_data.get("title"),
    }
    content = (post_data.get("title"), post_data.get("body"), post_data.get("comments"))

    task1 = asyncio.create_task(create_header_tag(root, meta_data))
    task2 = asyncio.create_task(create_body_tags(root, content, meta_data["author"]))
    await asyncio.gather(task1, task2)


def generate_tei_xml(post_data: PostData, media: Media) -> str:
    """The generate_tei_xml function generates TEI XML string.

    Args:
        post_data (PostData): the post data
        media (Media): a Taiwan social media name
    """

    if not post_data:
        raise ValueError("post_data cannot be empty")

    root = ET.Element("TEI.2")
    asyncio.run(create_tags(root, post_data, media))

    xml_str = minidom.parseString(ET.tostring(root)).childNodes[0]
    return xml_str.toprettyxml(indent="   ")



from .cleaners import (
    BodyCleaner,
    CommentsCleaner,
    TitleCleaner,
)
from .taggers import (
    BodyTagger,
    CommentsTagger,
    TitleTagger,
)

__all__ = [
    "TitleCleaner",
    "BodyCleaner",
    "CommentsCleaner",
    "TitleTagger",
    "BodyTagger",
    "CommentsTagger",
]



from datetime import datetime


def get_year(timestamp: str | int) -> str:
    """The get_year function converts `timestamp` to a year string.

    Args:
        timestamp (str | int): the timestamp to be converted
    Returns:
        a str
    """
    timestamp = int(timestamp) if isinstance(timestamp, str) else timestamp
    date_time = datetime.fromtimestamp(timestamp)
    return str(date_time.year)



import asyncio
import xml.etree.ElementTree as ET

from ckip2tei.ckip import segment
from ckip2tei.tei.utils import filter_body

from .taggers import (
    BodyCleaner,
    BodyTagger,
    CommentsCleaner,
    CommentsTagger,
    TitleCleaner,
    TitleTagger,
)

# --------------------------------------------------------------------
# helper functions


async def create_title_task(text_tag: ET.Element, data: str, author: str):
    """The create_title_task function handles the title tag creation."""
    segmented_data = await segment(data)

    if not segmented_data:
        return None

    title_tag = ET.SubElement(text_tag, "title", author=author)
    sentence_tag = ET.SubElement(title_tag, "s")
    tagger = TitleTagger(sentence_tag, TitleCleaner(segmented_data))
    await tagger.create()


async def create_body_task(text_tag, data, author):
    """The create_body_task function handles the body tag creation."""
    body_tag = ET.SubElement(text_tag, "body", author=author)
    segmented_data = await segment(filter_body(data))
    tagger = BodyTagger(body_tag, BodyCleaner(segmented_data))
    await tagger.create()


async def create_comments_task(text_tag, data):
    """The create_comments_task function handles the comment tags creation."""
    segmented_data = await segment(data)
    tagger = CommentsTagger(text_tag, CommentsCleaner(segmented_data))
    await tagger.create()


# --------------------------------------------------------------------
# public interface


async def create_body_tags(root: ET.Element, content: tuple, author: str):
    title, body, comments = content
    text_tag = ET.SubElement(root, "text")

    title_task = asyncio.create_task(create_title_task(text_tag, title, author))
    body_task = asyncio.create_task(create_body_task(text_tag, body, author))
    comments_task = asyncio.create_task(create_comments_task(text_tag, comments))
    await asyncio.gather(title_task, body_task, comments_task)



from dataclasses import dataclass

from .base import Cleaner

SegmentedSentences = list[list[tuple[str, str]]]


@dataclass(slots=True)
class TitleCleaner(Cleaner):
    """
    The TitleCleaner object cleans the title data output.
    """

    segmented_sentences: SegmentedSentences

    def clean(self):
        invalid_senntence = not self.segmented_sentences or self.segmented_sentences[0][
            0
        ][0].startswith("http")

        if invalid_senntence:
            return ""
        return self.segmented_sentences[0]


@dataclass(slots=True)
class BodyCleaner(Cleaner):
    """
    The BodyDataCleaner object cleans the body data output.
    """

    segmented_sentences: SegmentedSentences

    def clean(self):
        if not self.segmented_sentences:
            return ""

        return self.segmented_sentences


@dataclass(slots=True)
class CommentsCleaner(Cleaner):
    """
    The CommentsCleaner object cleans the comments data output.
    """

    segmented_sentences: SegmentedSentences

    def clean(self):
        if not self.segmented_sentences:
            return ""

        return self.segmented_sentences



from dataclasses import (
    dataclass,
    field,
)
from pathlib import Path
import pickle

from ckip2tei.config import (
    CKIP_DIR,
    CKIP_PATH,
    NLP_MODEL,
)


@dataclass(frozen=True, slots=True)
class CKIPClient:
    """
    The CKIPClient object connects to ckip drivers.
    """

    _ckip_path: str = field(init=False, default=CKIP_PATH)
    _nlp_model: str = field(init=False, default=NLP_MODEL)

    def __post_init__(self) -> None:
        self.on_ready()

    def on_ready(self) -> None:
        """The on_ready method initializes and caches the CKIP drivers."""
        has_path = Path(self._ckip_path).exists()

        if not has_path:
            from ckip_transformers.nlp import (
                CkipPosTagger,
                CkipWordSegmenter,
            )

            Path(CKIP_DIR).mkdir(parents=True, exist_ok=True)
            drivers = (
                CkipWordSegmenter(model=self._nlp_model),
                CkipPosTagger(model=self._nlp_model),
            )

            with open(rf"{self._ckip_path}", "wb") as file:
                pickle.dump(drivers, file)

    def connect(self) -> tuple:
        """The connect method connects to the ckip drivers.

        Returns:
            a tuple that contains CkipWordSegmenter and CkipPosTagger.
        """
        with open(self._ckip_path, "rb") as file:
            return pickle.load(file)



import asyncio

from ..model import CKIPClient

ws_driver, pos_driver = CKIPClient().connect()


async def pack_ws_pos(ws_pos_pair: tuple) -> list[tuple]:
    """The pack_ws_pos_sentece method packs a word and its part-of-speech to a pair.

    Args:
        ws_pos_pair (tuple): the pair of a word and its corresponding part-of-speech
    Returns:
        a list of tuples: [
            ('我', 'Nh'),
            ('喜歡', 'VK'),
            ('程式', 'Na')
        ]
    """
    sentence_ws, sentence_pos = ws_pos_pair
    assert len(sentence_ws) == len(sentence_pos)
    return list(zip(sentence_ws, sentence_pos))


async def transform(sentence: str | list[str]):
    """The transform method transforms the sentences in a list to word segmentation and
    part-of-speech results.
    """
    invalid_sentence = not sentence or all(value == "" for value in sentence)

    if invalid_sentence:
        return []

    data = [sentence.strip()] if isinstance(sentence, str) else sentence
    ws_pipeline = ws_driver(data, use_delim=True)
    pos_pipeline = pos_driver(ws_pipeline, use_delim=True)
    return await asyncio.gather(*map(pack_ws_pos, zip(ws_pipeline, pos_pipeline)))



import xml.etree.ElementTree as ET
from xml.sax.saxutils import escape

from .base import TeiTagger


class TitleTagger(TeiTagger):
    """
    The TitleTagger object creats the tei tags for the post title.
    """

    async def build_tag(self, tag_data: tuple[str, str]) -> None:
        if not tag_data:
            return None

        word, pos = tag_data
        ET.SubElement(self.tag, "w", type=pos).text = escape(word)


class BodyTagger(TeiTagger):
    """
    The BodyTagger object creats the tei tags for the post body.
    """

    async def build_tag(self, tag_data: list[tuple[str, str]]) -> None:
        sentence_tag = ET.SubElement(self.tag, "s")
        for word, pos in tag_data:
            ET.SubElement(sentence_tag, "w", type=pos).text = escape(word)


class CommentsTagger(TeiTagger):
    """
    The CommentsTagger object creats the tei tags for the post comments.
    """

    async def build_tag(self, tag_data: dict[str, list[tuple[str, str]]]):
        comments = tag_data.get("content")

        if not comments:
            return None

        author = tag_data.get("author", "匿名")
        comment_type = tag_data.get("type")
        comment_tag = ET.SubElement(
            self.tag, "comment", author=author, c_type=comment_type
        )
        sentence_tag = ET.SubElement(comment_tag, "s")

        for ws_pos_pair in comments[0]:
            word, pos = ws_pos_pair
            ET.SubElement(sentence_tag, "w", type=pos).text = escape(word)



import asyncio

from .transformer import transform

# --------------------------------------------------------------------
# helper functions


async def segment_comment(comment_data: dict[str, str]):
    """The segment_comment function segments the `content` value from `comment_data`.

    Args:
        comment_data (dict): a dict containing the comment data
    Returns:
        a dict: {
            'type': 'pos',
            'author': 'lixing',
            'content': [[
                ('我', 'Nh'),
                ('喜歡', 'VK'),
                ('程式', 'Na')
            ]],
            'order': '1',
        }
    """
    content = await transform(comment_data["content"])
    return {**comment_data, "content": content}


async def segment_comments(list_of_comments: list[dict[str, str]]):
    """The segment_comments function segments a list of comments."""
    return await asyncio.gather(*list(map(segment_comment, list_of_comments)))


# --------------------------------------------------------------------
# public interface


async def segment(data: str | list[str] | list[dict[str, str]]):
    """The segment function segments the `data` based on its type.

    Args:
        data (str | list): the post-related data (i.e title, body or comments)
    Returns:
        a list of dicts if the data refers to comments, a list of list of tuples
        otherwise.
    """
    is_comment_data = isinstance(data, list) and all(
        isinstance(value, dict) for value in data
    )

    if is_comment_data:
        return await segment_comments(data)

    return await transform(data)



from abc import (
    ABC,
    abstractmethod,
)


class Cleaner(ABC):
    """
    The Cleaner object cleans the segmented sentences output.
    """

    @abstractmethod
    def clean(self):
        pass
